@@ -18,7 +18,10 @@ use crate::{clarity::diagnostic::Diagnostic, repl::settings::InitialContract};
 use ansi_term::{Colour, Style};
 use std::collections::{BTreeMap, BTreeSet, HashMap, VecDeque};
 use std::fmt;
+use std::fs::{self, File};
+use std::io::Write;
 use std::num::ParseIntError;
+use std::path::PathBuf;
 use std::sync::{Arc, Mutex};
 
 #[cfg(feature = "cli")]
@@ -78,11 +81,7 @@ impl Session {
             session_id: 0,
             started_at: 0,
             is_interactive: false,
-            interpreter: ClarityInterpreter::new(
-                tx_sender,
-                settings.costs_version,
-                settings.analysis.clone(),
-            ),
+            interpreter: ClarityInterpreter::new(tx_sender, settings.repl_settings.clone()),
             asts: BTreeMap::new(),
             contracts: BTreeMap::new(),
             api_reference: build_api_reference(),
@@ -94,7 +93,8 @@ impl Session {
         }
     }
 
-    async fn retrieve_contract(
+    #[cfg(not(feature = "wasm"))]
+    fn retrieve_contract(
         &mut self,
         link: &InitialLink,
     ) -> Result<(String, Vec<QualifiedContractIdentifier>), String> {
@@ -102,37 +102,70 @@ impl Session {
         let components: Vec<&str> = contract_id.split('.').collect();
         let contract_deployer = components.first().expect("");
         let contract_name = components.last().expect("");
-        let stacks_node_addr = match &link.stacks_node_addr {
-            Some(addr) => addr.clone(),
-            None => {
-                if contract_id.starts_with("SP") {
-                    "https://stacks-node-api.mainnet.stacks.co".to_string()
-                } else {
-                    "https://stacks-node-api.testnet.stacks.co".to_string()
-                }
+
+        let mut contract_source = None;
+        if let Some(ref cache_path) = link.cache {
+            let mut file_path = PathBuf::from(cache_path);
+            file_path.push(format!("{}.clar", contract_id));
+            if let Ok(data) = fs::read_to_string(file_path) {
+                contract_source = Some(data);
             }
+        }
+
+        let code = if contract_source.is_none() {
+            let stacks_node_addr = match &link.stacks_node_addr {
+                Some(addr) => addr.clone(),
+                None => {
+                    if contract_id.starts_with("SP") {
+                        "https://stacks-node-api.mainnet.stacks.co".to_string()
+                    } else {
+                        "https://stacks-node-api.testnet.stacks.co".to_string()
+                    }
+                }
+            };
+
+            let request_url = format!(
+                "{host}/v2/contracts/source/{addr}/{name}?proof=0",
+                host = stacks_node_addr,
+                addr = contract_deployer,
+                name = contract_name
+            );
+
+            let rt = tokio::runtime::Runtime::new().unwrap();
+            let response = rt.block_on(async { fetch_contract(request_url).await });
+            response.source.to_string()
+        } else {
+            contract_source.unwrap()
         };
 
-        let request_url = format!(
-            "{host}/v2/contracts/source/{addr}/{name}?proof=0",
-            host = stacks_node_addr,
-            addr = contract_deployer,
-            name = contract_name
-        );
+        if self.settings.disk_cache_enabled {
+            if let Some(ref cache_path) = link.cache {
+                let mut file_path = PathBuf::from(cache_path);
+                let _ = fs::create_dir_all(&file_path);
+                file_path.push(format!("{}.clar", contract_id));
 
-        let response = fetch_contract(request_url).await;
+                if let Ok(ref mut file) = File::create(file_path) {
+                    let _ = file.write_all(code.as_bytes());
+                }
+            }
+        }
 
-        let code = response.source.to_string();
         let deps = self
             .interpreter
-            .detect_dependencies(contract_id.to_string(), code.clone())
+            .detect_dependencies(
+                contract_id.to_string(),
+                code.clone(),
+                self.settings.repl_settings.parser_version,
+            )
             .unwrap();
         Ok((code, deps))
     }
 
-    pub async fn resolve_link(
+    #[cfg(not(feature = "wasm"))]
+    pub fn resolve_link(
         &mut self,
         link: &InitialLink,
+        retrieved: &mut BTreeSet<String>,
     ) -> Result<Vec<(String, String, Vec<String>)>, String> {
         let mut resolved_link = Vec::new();
 
@@ -142,10 +175,11 @@ impl Session {
         queue.push_front(link.clone());
 
         while let Some(initial_link) = queue.pop_front() {
+            if retrieved.contains(&initial_link.contract_id) {
+                continue;
+            }
+
             let contract_id = &initial_link.contract_id;
-            let components: Vec<&str> = contract_id.split('.').collect();
-            let contract_deployer = components.first().expect("");
-            let contract_name = components.last().expect("");
 
             // Extract principal from contract_id
             let (contract_code, deps) = match handled.get(contract_id) {
@@ -153,7 +187,6 @@ impl Session {
                 None => {
                     let (contract_code, deps) = self
                         .retrieve_contract(&initial_link)
-                        .await
                         .expect("Unable to get contract");
                     handled.insert(contract_id.to_string(), contract_code.clone());
                     (contract_code, deps)
@@ -188,7 +221,7 @@ impl Session {
     #[cfg(not(feature = "wasm"))]
     pub fn start(&mut self) -> Result<(String, Vec<(ContractAnalysis, String, String)>), String> {
         let mut output_err = Vec::<String>::new();
-        let mut output = vec![];
+        let mut output = Vec::<String>::new();
         let mut contracts = vec![];
 
         if !self.settings.include_boot_contracts.is_empty() {
@@ -210,49 +243,6 @@ impl Session {
             self.interpreter.set_tx_sender(default_tx_sender);
         }
 
-        let mut linked_contracts = Vec::new();
-
-        if self.settings.initial_links.len() > 0 {
-            let initial_links = self.settings.initial_links.clone();
-            let default_tx_sender = self.interpreter.get_tx_sender();
-
-            let mut indexed = BTreeSet::new();
-
-            let rt = tokio::runtime::Runtime::new().unwrap();
-            for link in initial_links.iter() {
-                let contracts = rt.block_on(async { self.resolve_link(link).await.unwrap() });
-                for (contract_id, code, _) in contracts.into_iter() {
-                    if !indexed.contains(&contract_id) {
-                        indexed.insert(contract_id.clone());
-                        linked_contracts.push((contract_id, code));
-                    }
-                }
-            }
-            for (contract_id, code) in linked_contracts.iter() {
-                let components: Vec<&str> = contract_id.split('.').collect();
-                let contract_deployer = components.first().expect("");
-                let contract_name = components.last().expect("");
-
-                let deployer = {
-                    PrincipalData::parse_standard_principal(&contract_deployer)
-                        .expect("Unable to parse deployer's address")
-                };
-
-                self.interpreter.set_tx_sender(deployer);
-                match self.formatted_interpretation(
-                    code.to_string(),
-                    Some(contract_name.to_string()),
-                    true,
-                    None,
-                ) {
-                    Ok((mut output, _)) => output_err.append(&mut output),
-                    Err(ref mut result) => output_err.append(result),
-                };
-            }
-
-            self.interpreter.set_tx_sender(default_tx_sender);
-        }
-
         if self.settings.initial_accounts.len() > 0 {
             let mut initial_accounts = self.settings.initial_accounts.clone();
             for account in initial_accounts.drain(..) {
@@ -274,6 +264,101 @@ impl Session {
             }
         }
 
+        if !self.settings.lazy_initial_contracts_interpretation {
+            match self.interpret_initial_contracts() {
+                Ok((ref mut res, ref mut initial_contracts)) => {
+                    if self.is_interactive {
+                        // If the session is interactive (clarinet console, usr/bin/clarity-repl)
+                        // we will display the contracts + genesis asset map.
+                        if !self.settings.initial_contracts.is_empty() {
+                            output.push(blue!("Contracts"));
+                            self.get_contracts(&mut output);
+                        }
+
+                        if self.settings.initial_accounts.len() > 0 {
+                            output.push(blue!("Initialized balances"));
+                            self.get_accounts(&mut output);
+                        }
+                    }
+                    contracts.append(initial_contracts);
+                }
+                Err(ref mut res) => {
+                    output_err.append(res);
+                }
+            };
+        }
+
+        match output_err.len() {
+            0 => Ok((output.join("\n"), contracts)),
+            _ => Err(output_err.join("\n")),
+        }
+    }
+
+    #[cfg(not(feature = "wasm"))]
+    fn handle_requirements(&mut self) -> Result<Vec<String>, Vec<String>> {
+        let mut output_err = vec![];
+        let mut output = vec![];
+
+        let mut linked_contracts = Vec::new();
+
+        if self.settings.initial_links.len() > 0 {
+            let initial_links = self.settings.initial_links.clone();
+            let default_tx_sender = self.interpreter.get_tx_sender();
+
+            let mut retrieved = BTreeSet::new();
+
+            for link in initial_links.iter() {
+                if retrieved.contains(&link.contract_id) {
+                    continue;
+                }
+                let contracts = self.resolve_link(link, &mut retrieved).unwrap();
+                for (contract_id, code, _) in contracts.into_iter() {
+                    if !retrieved.contains(&contract_id) {
+                        retrieved.insert(contract_id.clone());
+                        linked_contracts.push((contract_id, code));
+                    }
+                }
+            }
+            for (contract_id, code) in linked_contracts.iter() {
+                let components: Vec<&str> = contract_id.split('.').collect();
+                let contract_deployer = components.first().expect("");
+                let contract_name = components.last().expect("");
+
+                let deployer = {
+                    PrincipalData::parse_standard_principal(&contract_deployer)
+                        .expect("Unable to parse deployer's address")
+                };
+
+                self.interpreter.set_tx_sender(deployer);
+                match self.formatted_interpretation(
+                    code.to_string(),
+                    Some(contract_name.to_string()),
+                    true,
+                    None,
+                ) {
+                    Ok((mut logs, _)) => output.append(&mut logs),
+                    Err(ref mut result) => {
+                        output_err.append(result);
+                        break;
+                    }
+                };
+            }
+
+            self.interpreter.set_tx_sender(default_tx_sender);
+        }
+        if output_err.len() > 0 {
+            return Err(output_err);
+        }
+        Ok(output)
+    }
+
+    fn handle_initial_contracts(
+        &mut self,
+    ) -> Result<(Vec<String>, Vec<(ContractAnalysis, String, String)>), Vec<String>> {
+        let mut output_err = vec![];
+        let mut output = vec![];
+
+        let mut contracts = vec![];
         if self.settings.initial_contracts.len() > 0 {
             let mut initial_contracts = self.settings.initial_contracts.clone();
             let default_tx_sender = self.interpreter.get_tx_sender();
@@ -311,44 +396,26 @@ impl Session {
             }
             self.interpreter.set_tx_sender(default_tx_sender);
         }
-
-        for (contract_id, code) in linked_contracts.into_iter() {
-            let components: Vec<&str> = contract_id.split('.').collect();
-            let contract_deployer = components.first().expect("");
-            let contract_name = components.last().expect("");
-
-            let deployer = {
-                PrincipalData::parse_standard_principal(&contract_deployer)
-                    .expect("Unable to parse deployer's address")
-            };
-            self.settings.initial_contracts.push(InitialContract {
-                code: code.to_string(),
-                path: "".into(),
-                name: Some(contract_id.to_string()),
-                deployer: Some(deployer.to_string()),
-            });
+        if output_err.len() > 0 {
+            return Err(output_err);
         }
+        Ok((output, contracts))
+    }
 
-        if self.is_interactive {
-            // If the session is interactive (clarinet console, usr/bin/clarity-repl)
-            // we will display the contracts + genesis asset map.
-            if !self.settings.initial_contracts.is_empty() {
-                output.push(blue!("Contracts"));
-                self.get_contracts(&mut output);
-            }
+    #[cfg(not(feature = "wasm"))]
+    pub fn interpret_initial_contracts(
+        &mut self,
+    ) -> Result<(Vec<String>, Vec<(ContractAnalysis, String, String)>), Vec<String>> {
+        if self.initial_contracts_analysis.is_empty() {
+            let output = self.handle_requirements()?;
+            let (output, contracts) = self.handle_initial_contracts()?;
 
-            if self.settings.initial_accounts.len() > 0 {
-                output.push(blue!("Initialized balances"));
-                self.get_accounts(&mut output);
-            }
-        }
+            self.initial_contracts_analysis
+                .append(&mut contracts.clone());
 
-        self.initial_contracts_analysis
-            .append(&mut contracts.clone());
-
-        match output_err.len() {
-            0 => Ok((output.join("\n"), contracts)),
-            _ => Err(output_err.join("\n")),
+            Ok((output, contracts))
+        } else {
+            Err(vec!["Initial contracts already interpreted".into()])
         }
     }
 
@@ -431,64 +498,10 @@ impl Session {
             self.interpreter.set_tx_sender(default_tx_sender);
         }
 
-        let mut linked_contracts = Vec::new();
-
-        if self.settings.initial_links.len() > 0 {
-            let initial_links = self.settings.initial_links.clone();
-            let default_tx_sender = self.interpreter.get_tx_sender();
-
-            let mut indexed = BTreeSet::new();
-            for link in initial_links.iter() {
-                let contracts = self.resolve_link(link).await.unwrap();
-                for (contract_id, code, _) in contracts.into_iter() {
-                    if !indexed.contains(&contract_id) {
-                        indexed.insert(contract_id.clone());
-                        linked_contracts.push((contract_id, code));
-                    }
-                }
-            }
-            for (contract_id, code) in linked_contracts.iter() {
-                let components: Vec<&str> = contract_id.split('.').collect();
-                let contract_deployer = components.first().expect("");
-                let contract_name = components.last().expect("");
-
-                let deployer = {
-                    PrincipalData::parse_standard_principal(&contract_deployer)
-                        .expect("Unable to parse deployer's address")
-                };
-
-                self.interpreter.set_tx_sender(deployer);
-                match self.formatted_interpretation(
-                    code.to_string(),
-                    Some(contract_name.to_string()),
-                    true,
-                    None,
-                ) {
-                    Ok((ref mut res_output, _)) => output.append(res_output),
-                    Err(ref mut result) => output.append(result),
-                };
-            }
-
-            self.interpreter.set_tx_sender(default_tx_sender);
-            self.get_contracts(&mut output);
-        }
-
-        for (contract_id, code) in linked_contracts.iter() {
-            let components: Vec<&str> = contract_id.split('.').collect();
-            let contract_deployer = components.first().expect("");
-            let contract_name = components.last().expect("");
-
-            let deployer = {
-                PrincipalData::parse_standard_principal(&contract_deployer)
-                    .expect("Unable to parse deployer's address")
-            };
-            self.settings.initial_contracts.push(InitialContract {
-                code: code.to_string(),
-                path: "".into(),
-                name: Some(contract_id.to_string()),
-                deployer: Some(deployer.to_string()),
-            });
-        }
+        // The cost of maintaining the "requirements" / "links" feature on WASM builds
+        // is pretty high, and the amount of code duplicated is very important.
+        // We will timeshift through git and restore this feature if we
+        // can identify a concrete use case in the future
 
         if !self.settings.initial_contracts.is_empty() {
             output.push(blue!("Contracts"));
@@ -576,9 +589,9 @@ impl Session {
                 }
                 Ok((output, result))
             }
-            Err((_, diagnostic, _)) => {
-                if let Some(diagnostic) = diagnostic {
-                    output.append(&mut diagnostic.output(&contract_name, &formatted_lines));
+            Err(diagnostics) => {
+                for d in diagnostics {
+                    output.append(&mut d.output(&contract_name, &formatted_lines));
                 }
                 Err(output)
             }
@@ -592,7 +605,7 @@ impl Session {
         args: &Vec<String>,
         sender: &str,
         test_name: String,
-    ) -> Result<ExecutionResult, (String, Option<Diagnostic>, Option<Error>)> {
+    ) -> Result<ExecutionResult, Vec<Diagnostic>> {
         let initial_tx_sender = self.get_tx_sender();
         // Kludge for handling fully qualified contract_id vs sugared syntax
         let first_char = contract.chars().next().unwrap();
@@ -630,7 +643,7 @@ impl Session {
         name: Option<String>,
         cost_track: bool,
         test_name: Option<String>,
-    ) -> Result<ExecutionResult, (String, Option<Diagnostic>, Option<Error>)> {
+    ) -> Result<ExecutionResult, Vec<Diagnostic>> {
         let (contract_name, is_tx) = match name {
             Some(name) => (name, false),
             None => (format!("contract-{}", self.contracts.len()), true),
@@ -851,10 +864,13 @@ impl Session {
                 }
                 green!(s)
             }
-            Err((_, Some(diagnostic), Some(e))) => red!(format!("{}: {}", diagnostic.message, e)),
-            Err((_, Some(diagnostic), None)) => red!(format!("{}", diagnostic.message)),
-            Err((_, None, Some(e))) => red!(format!("{}", e)),
-            _ => panic!("Result of interpret has neither diagnostic or error"),
+            Err(diagnostics) => {
+                let lines: Vec<String> = snippet.split('\n').map(|s| s.to_string()).collect();
+                for d in diagnostics {
+                    output.append(&mut d.output(&"encode".to_string(), &lines));
+                }
+                red!("encoding failed")
+            }
         };
         output.push(value);
     }
@@ -1214,15 +1230,22 @@ mod tests {
         let mut session = Session::new(SessionSettings::default());
         let mut output: Vec<String> = Vec::new();
         session.encode(&mut output, "::encode { foo false }");
-        assert_eq!(output.len(), 1);
         assert_eq!(
             output[0],
-            red!("Tuple literal construction expects a colon at index 1")
+            format!(
+                "encode:1:7: {}: expected ':' after key in tuple",
+                red!("error")
+            )
         );
 
         session.encode(&mut output, "::encode (foo 1)");
-        assert_eq!(output.len(), 2);
-        assert_eq!(output[1], red!("use of unresolved function 'foo'"));
+        assert_eq!(
+            output[4],
+            format!(
+                "encode:1:1: {}: use of unresolved function 'foo'",
+                red!("error")
+            )
+        );
     }
 
     #[test]
@@ -1269,7 +1292,7 @@ mod tests {
     fn evaluate_at_block() {
         let mut settings = SessionSettings::default();
         settings.include_boot_contracts = vec!["costs-v1".into()];
-        settings.costs_version = 1;
+        settings.repl_settings.costs_version = 1;
 
         let mut session = Session::new(settings);
         session.start().expect("session could not start");
